// $Date:: 2010-03-27 23:23:56 +0900#$
// $Rev: 27 $

#include "Smoother.h"

//#define PRINTTIME
#ifdef PRINTTIME
#include <stdio.h>
#include <mmsystem.h>
#pragma comment ( lib , "winmm.lib" )
static __int32 tcount = 0 ;
#endif

/**************************************************************************************************/
/* Smoother                                                                                      */
/**************************************************************************************************/
//==============================================================================
// constructor
//==============================================================================
Smoother::Smoother(PClip _child, int strength, int threshold, int type, IScriptEnvironment* env) : GenericVideoFilter(_child)
{
    this->Height    = this->vi.height ;
    this->Width     = this->vi.width ;
    
    // mCYxB0`20 ͈̔͂ŏB
    this->Strength  = strength ;

    // GbWo臒lB0`765 ͈̔͂ŏB
    this->Threshold = threshold ;

    // g̒lvZ^CvB0,1,2 ŏB
    this->Type = type ;

    // p[^͈̓`FbN
    if ( this->Strength < 0 || this->Strength > 20 )
        env->ThrowError("Smoother: strength must be 0 to 20.");

    if ( this->Threshold < 0 || this->Threshold > 765 )
        env->ThrowError("Smoother: threshold must be 0 to 765.");

    if ( this->Type < 0 || this->Type > 4 )
        env->ThrowError("Smoother: type must be 0 to 20.");

    this->out = (BYTE*) malloc ( this->child->GetFrame ( 0 , env )->GetPitch() * this->Height * sizeof (BYTE) ) ;
}

//==============================================================================
// destructor
//==============================================================================
Smoother::~Smoother()
{
    free ( this->out ) ;
#ifdef PRINTTIME
    char strtime [ 256 ] ;
    _snprintf ( strtime , sizeof (strtime) , "%f\r\n" , ( (float) tcount ) / 1000 ) ;
    OutputDebugString ( strtime ) ;
#endif // PRINTTIME
}

//==============================================================================
// GetFrame
//==============================================================================
PVideoFrame Smoother::GetFrame (int n, IScriptEnvironment* env)
{
    PVideoFrame src = this->child->GetFrame ( n , env ) ;
    env->MakeWritable ( & src ) ;

    BYTE* pSrcY = src->GetWritePtr () ;
    const int nSrcYPitch = src->GetPitch();

#ifdef PRINTTIME
    tcount -= timeGetTime () ;
#endif // PRINTTIME

    memcpy ( this->out , pSrcY , nSrcYPitch * this->Height * sizeof (BYTE) ) ;

    for ( int i = 1 ; i < this->Height - 1 ; i ++ )
    {
        for ( int j = 1 ; j < this->Width - 1 ; j ++ )
        {
            BYTE v [ 3 ][ 4 ] ;
            memcpy ( v [ 0 ] , pSrcY + nSrcYPitch * ( i - 1 ) + j - 1 , 4 ) ;
            memcpy ( v [ 1 ] , pSrcY + nSrcYPitch * ( i + 0 ) + j - 1 , 4 ) ;
            memcpy ( v [ 2 ] , pSrcY + nSrcYPitch * ( i + 1 ) + j - 1 , 4 ) ;

            __int16 avg ;

            switch ( this->Type )
            {
            case 1 :
                avg = 
                    v [ 0 ][ 1 ] + v [ 1 ][ 0 ] + v [ 1 ][ 1 ] + v [ 1 ][ 2 ] + v [ 2 ][ 1 ] -
                    __max ( __max ( __max ( __max (
                        v [ 0 ][ 1 ] , v [ 1 ][ 0 ] ) , v [ 1 ][ 1 ] ) , v [ 1 ][ 2 ] ) , v [ 2 ][ 1 ] ) -
                    __min ( __min ( __min ( __min (
                        v [ 0 ][ 1 ] , v [ 1 ][ 0 ] ) , v [ 1 ][ 1 ] ) , v [ 1 ][ 2 ] ) , v [ 2 ][ 1 ] ) ;
                break ;

            case 2 :
                avg = (
                    v [ 0 ][ 0 ] + v [ 0 ][ 1 ] + v [ 0 ][ 2 ] +
                    v [ 1 ][ 0 ] + v [ 1 ][ 1 ] + v [ 1 ][ 2 ] +
                    v [ 2 ][ 0 ] + v [ 2 ][ 1 ] + v [ 2 ][ 2 ] ) / 3 ;
                break ;

            default :
                avg = (__int16) v [ 1 ][ 1 ] * 3 ;
            }

            __int16 a [ 2 ] = {
                __max ( abs ( v [ 0 ][ 0 ] + v [ 0 ][ 1 ] + v [ 0 ][ 2 ] - avg ) , abs ( v [ 2 ][ 2 ] + v [ 2 ][ 1 ] + v [ 2 ][ 0 ] - avg ) ) ,
                __max ( abs ( v [ 0 ][ 2 ] + v [ 1 ][ 2 ] + v [ 2 ][ 2 ] - avg ) , abs ( v [ 2 ][ 0 ] + v [ 1 ][ 0 ] + v [ 0 ][ 0 ] - avg ) )
            };

            __int16 t ;
            __int16 aa ;
            if ( a [ 0 ] > a [ 1 ] )
            {
                aa = a [ 0 ] ;
                t = v [ 1 ][ 2 ] + v [ 1 ][ 0 ] ;
            }
            else
            {
                aa = a [ 1 ] ;
                t = v [ 2 ][ 1 ] + v [ 0 ][ 1 ] ;
            }
 
            if ( aa > this->Threshold )
            {
                this->out [ nSrcYPitch * i + j ] = ( t * this->Strength + v [ 1 ][ 1 ] * ( 40 - 2 * this->Strength ) ) / 40 ;
            }
        }
    }

    memcpy ( pSrcY , this->out , nSrcYPitch * this->Height * sizeof (BYTE) ) ;

#ifdef PRINTTIME
    tcount += timeGetTime () ;
#endif // PRINTTIME
    return src ;
}
